/***************************************************************************
                          hexapod.cpp  -  description
                             -------------------
    begin                : Sun Mar 17 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "hexapod.h"      // import constants for bodies
#include <iostream>

using std::cout;
using std::endl;

Hexapod::Hexapod(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z) :
	Creature(world,space,x,y,z), numBody(3), numGeom(3), numJoint(2)		// Build Creature
{
	// Body Setup
	TORSO[0] = 0.4; TORSO[1] = 0.8; TORSO[2] = 0.2;
	LEG[0] = 0.1; LEG[1] = 0.07; LEG[2] = 0.28; GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	//geomGroup = dCreateGeomGroup(space);

	// constants (Mass)
	const dReal MTORSO=2.8;
	const dReal MLEG=0.5;

	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	// right legs
	dBodySetPosition (body[1],x+TORSO[0]/3,y-TORSO[1]/2,z-TORSO[2]/2);	//right forward leg
	// left legs
	dBodySetPosition (body[2],x-TORSO[0]/3,y-TORSO[1]/2,z-TORSO[2]/2);	//left forward leg
	
	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[0],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[0]);

	// legs
	dMassSetBox (&m,1,LEG[0],LEG[1],LEG[2]);	// legs
	dMassAdjust (&m,MLEG);	//  leg weight
	for (int i=1;i<=2;i++) {
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG[0],LEG[1],LEG[2]);
		dGeomSetBody (geom[i],body[i]);
	}


	/********************************** CREATE JOINTS **********************************/
	// Leg Sockets 1
	for (int i=0;i<=1;i++) {
		joint[i] = dJointCreateHinge(world,jGroup);
		dJointAttach(joint[i],body[0],body[i+1]);
	}
	dJointSetHingeAnchor(joint[0],x+TORSO[0]/2,y-TORSO[1]/2,z-TORSO[2]/2+LEG[2]/2);
	dJointSetHingeAnchor(joint[1],x-TORSO[0]/2,y-TORSO[1]/2,z-TORSO[2]/2+LEG[2]/2);
	for (int i=0;i<=1;i++)
	{
		dJointSetHingeAxis(joint[i],1,0,0);
		dJointSetHingeParam(joint[i],dParamLoStop,-0.5);
		dJointSetHingeParam(joint[i],dParamHiStop,1.5);
		dJointSetHingeParam(joint[i],dParamStopERP,0.2);
		dJointSetHingeParam(joint[i],dParamStopCFM,0.05);
		dJointSetHingeParam(joint[i],dParamBounce,0.3);
		dJointSetHingeParam(joint[i],dParamFudgeFactor,0.5);
		dJointSetHingeParam(joint[i],dParamVel,0);
		dJointSetHingeParam(joint[i],dParamFMax,2);
	}
}

Hexapod::~Hexapod()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
//	dGeomDestroy(geomGroup);
	// destroy Geometries
/*	for (int i=0;i<5;i++) {
		dGeomGroupRemove(geomGroup,geom[i]);
	} */
	for (int i=0;i<numGeom;i++) {
		//dGeomGroupRemove(geomGroup,geom[i]);
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}


void Hexapod::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawBox(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),TORSO);
	if (selected)
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// legs
	for (int i=1;i<=2;i++)	
		dsDrawBox(dBodyGetPosition(body[i]),dBodyGetRotation(body[i]),LEG);
}

void Hexapod::setMuscles(const float * const inputs)
{
	// right Leg Socket
	dJointSetHingeParam(joint[0],dParamVel,(dReal) inputs[0]*2.5);
	dJointSetHingeParam(joint[1],dParamVel,(dReal) inputs[1]*2.5);
}

void Hexapod::getInputs(float * inputs) const
{
	// position inputs (position & rotation of head)
	const dReal * angle = dBodyGetRotation(body[0]);

	inputs[0] = (float)dJointGetHingeAngleRate(joint[0]);
	inputs[1] = (float)dJointGetHingeAngleRate(joint[1]);

	// direction sensor
	inputs[2] = (float)angle[1];

	// foot touch sensor
	if (r_foot_sens)
		inputs[3] = 1.0;
	else	
		inputs[3] = -1.0;
	if (l_foot_sens)
		inputs[4] = 1.0;
	else
		inputs[4] = -1.0;
}

void Hexapod::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[1]) && (o2==ground)) || ((o2==geom[1])&&(o1==ground)))
		r_foot_sens=true;
	else
		r_foot_sens=false;
	if (((o1==geom[2]) && (o2==ground)) || ((o2==geom[2])&&(o1==ground)))
		l_foot_sens=true;
	else
		l_foot_sens=false;
}

void Hexapod::shutoff(void)
{
	dJointSetHingeParam(joint[0],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[0],dParamFMax,(dReal)0);
	dJointSetHingeParam(joint[1],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[1],dParamFMax,(dReal)0);	
}


void Hexapod::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	// dummy function
}

const dReal *Hexapod::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

const dReal *Hexapod::getBodyRot(void) const
{
	return dBodyGetRotation(body[0]);
}

const float Hexapod::getFitness(void) const
{
	const dReal *pos = getHeadPos();
	float fitness = -pos[1]*9 - pow(fabs(getStartX()-pos[0]),4);
  if (fitness > 0)
		return fitness;
	else
		return 0;		// y coordinate.
}

void Hexapod::randomForce(void) const
{
    dReal f[3];
    f[0] = 22.5 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 42.5 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = .5 * (-0.5+1.0/RAND_MAX * rand());
    dBodyAddForce(body[0], f[0], f[1], f[2]);
}
void Hexapod::randomTorque(void) const
{
    dReal f[3];
    f[0] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
	  dBodyAddTorque(body[0], f[0], f[1], f[2]);
}

bool Hexapod::hasFallen(void) const
{
	return false;
}

void Hexapod::printStats(void) const
{
	const dReal *pos = dBodyGetPosition(body[0]);
	const dReal *rot = dBodyGetRotation(body[0]);

	cout << "\nPosition:";
	cout << "X: " << pos[0] << " ";
	cout << "Y: " << pos[1] << " ";
	cout << "Z: " << pos[2] << endl;
	cout << "Rotation:" << endl;
	cout << "XR: " << rot[0] << " ";
	cout << "YR: " << rot[1] << " ";
	cout << "ZR: " << rot[2] << endl;
}

void Hexapod::clearSensor(void)
{
  r_foot_sens=false;
  l_foot_sens=false;
}

