/***************************************************************************
                          newbox.cpp  -  description
                             -------------------
    begin                : Thu Feb 12 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "newbox.h"      // import constants for bodies
#include <iostream>

using std::cout;
using std::endl;

template class Organism<float>;
template class Organism<double>;
template class NewBox<float>;
template class NewBox<double>;

template <class T>
NewBox<T>::NewBox(dWorldID world, dSpaceID space, dVector3 position, bool enabled) :
Organism<T>(world,space,position,enabled),
numBody(7),
numGeom(7),
numJoint(6)
{
	// Brain Setup
	m_Brain = new NeuralNet<T>(9,10,6,10,1);
	setResponseTime(5);		// calc response every 5 iterations
	constructBody(world,space,position);
}

template <class T>
NewBox<T>::~NewBox()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	for (int i=0;i<numGeom;i++) {
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete m_Brain;
	delete [] body;
	delete [] geom;
	delete [] joint;
}

template <class T>
NewBox<T>::NewBox(const NewBox<T> &right) :
Organism<T>(right)
{
	numBody = right.numBody;
	numGeom = right.numGeom;
	numJoint = right.numJoint;
	NeuralNet<T> *temp = dynamic_cast<NeuralNet<T> *>(right.m_Brain);
	m_Brain = new NeuralNet<T>(*temp);
	constructBody(right.m_WorldID, right.m_SpaceID, right.m_StartPosition);
}

template <class T>
void NewBox<T>::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawBox(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),TORSO);
	if (selected || isSelected())
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// legs
	dsDrawBox(dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),LEG1);
	dsDrawBox(dBodyGetPosition(body[2]),dBodyGetRotation(body[2]),LEG1);
	dsDrawBox(dBodyGetPosition(body[3]),dBodyGetRotation(body[3]),LEG2);
	dsDrawBox(dBodyGetPosition(body[4]),dBodyGetRotation(body[4]),LEG2);
	dsDrawBox(dBodyGetPosition(body[5]),dBodyGetRotation(body[5]),LEG3);
	dsDrawBox(dBodyGetPosition(body[6]),dBodyGetRotation(body[6]),LEG3);
}

template <class T>
void NewBox<T>::activateMotors(const vector<T> &inputs)
{
	// right Leg Socket
	dJointSetHingeParam(joint[0],dParamVel,(dReal) inputs[0]*2.55);
	dJointSetHingeParam(joint[0],dParamFMax,(dReal)2.55);
	// left leg socket
	dJointSetHingeParam(joint[1],dParamVel,(dReal) inputs[1]*2.55);
	dJointSetHingeParam(joint[1],dParamFMax,(dReal)2.55);
	// right foot
	dJointSetHingeParam(joint[2],dParamVel,(dReal)inputs[2]*3.2);
	dJointSetHingeParam(joint[2],dParamFMax,(dReal)3.7);
	// left foot
	dJointSetHingeParam(joint[3],dParamVel,(dReal)inputs[3]*3.2);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)3.7);
	// right foot
	dJointSetHingeParam(joint[4],dParamVel,(dReal)inputs[4]*1.7);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)2.5);
	// left foot
	dJointSetHingeParam(joint[5],dParamVel,(dReal)inputs[5]*1.7);
	dJointSetHingeParam(joint[5],dParamFMax,(dReal)2.5);
}

template <class T>
vector<T> NewBox<T>::getSensoryData() const
{
	vector<T> outputs;
	// position inputs (position & rotation of head)
	const dReal * angle = dBodyGetRotation(body[0]);
	// get right leg socket angles (3)
	outputs.push_back(dJointGetHingeAngleRate(joint[0]));
	outputs.push_back(dJointGetHingeAngleRate(joint[1]));
	// get left leg socket angles (3)
	// get right foot angle
	outputs.push_back(dJointGetHingeAngleRate(joint[2]));
	// get left foot angle
	outputs.push_back(dJointGetHingeAngleRate(joint[3]));
	// get right foot angle
	outputs.push_back(dJointGetHingeAngleRate(joint[4]));
	// get left foot angle
	outputs.push_back(dJointGetHingeAngleRate(joint[5]));
	//outputs.push_back(angle[0]);
	outputs.push_back(angle[1]);
	//outputs.push_back(angle[2]);

	if (r_foot_sens)
		outputs.push_back(1.0);
	else
		outputs.push_back(-1.0);
	if (l_foot_sens)
		outputs.push_back(1.0);
	else
		outputs.push_back(-1.0);
  
	return outputs;
}

template <class T>
void NewBox<T>::groundContact(dGeomID ground)
{
	const int N = 5;	// approximate 10 contacts
	dContact contact[N];
	r_foot_sens = false;
	l_foot_sens = false;

	int n = dCollide (geom[5],ground,N,&contact[0].geom,sizeof(dContact));
  // contact!
	if (n > 0) {
		r_foot_sens=true;
	}

	n = dCollide (geom[6],ground,N,&contact[0].geom,sizeof(dContact));
	if (n > 0) {
		l_foot_sens=true;
  }
}

template <class T>
void NewBox<T>::reset()
{
	// reset the foot sensors
  r_foot_sens=false;
  l_foot_sens=false;
  
  NeuralNet<T> *brain_ptr = dynamic_cast<NeuralNet<T> *>(m_Brain);
  int output_size = brain_ptr->getOutputSize();
  vector<T> motor_outputs;
  for (int i=0;i<output_size;++i)
  {
		motor_outputs.push_back(0);
  }
  // reset the motors to zero
  activateMotors(motor_outputs);  
}

template <class T>
const dReal *NewBox<T>::getCurrentPosition() const
{
	return dBodyGetPosition(body[0]);
}

template <class T>
const dReal *NewBox<T>::getBodyRot() const
{
	return dBodyGetRotation(body[0]);
}

template <class T>
const double NewBox<T>::calcFitness() const
{
	const dReal *pos = getCurrentPosition();
  double result = -pos[1]*9 - pow(fabs(m_StartPosition[0]-pos[0]),4);

	if (result > 0)
		return result;
	else
		return 0;		// y coordinate.
}

template <class T>
void NewBox<T>::printStats(void) const
{
	const dReal *pos = dBodyGetPosition(body[0]);
	const dReal *rot = dBodyGetRotation(body[0]);

	cout << "\nPosition:";
	cout << "X: " << pos[0] << " ";
	cout << "Y: " << pos[1] << " ";
	cout << "Z: " << pos[2] << endl;
	cout << "Rotation:" << endl;
	cout << "XR: " << rot[0] << " ";
	cout << "YR: " << rot[1] << " ";
	cout << "ZR: " << rot[2] << endl;
	cout << "Feet:" << endl;
	cout << "Right: " << r_foot_sens << " ";
	cout << "Left: " << l_foot_sens << " " << endl;
}

template <class T>
vector<T> NewBox<T>::getBrainOutput(vector<T> &input)
{
	return(m_Brain->activate(input));
}

template <class T>
Organism<T> *NewBox<T>::operator+(const Organism<T> &right)
{
  vector<T> leftChromo = m_Brain->getWeights();
  vector<T> rightChromo = dynamic_cast<const NewBox<T> *>(&right)->m_Brain->getWeights();
  
	int length = m_Brain->getGeneSize();
  vector<T> chromosome;

  // Chance of Crossover
	if (RandFloat() <= CROSSOVER)
	{
	 	int crosspoint = (RandInt(0,length-1));

	  // Determine Crossover Point 
	  for (int i=0;i<length;i++)
		{
			if (i <= crosspoint)
				chromosome.push_back(leftChromo[i]);
			else
				chromosome.push_back(rightChromo[i]);
		}
	}
	else // just copy
	{
		if (RandFloat() < 0.50) {
			for (int i=0;i<length;i++)
				chromosome.push_back(leftChromo[i]);
		}
		else {
			for (int i=0;i<length;i++)
				chromosome.push_back(rightChromo[i]);
		}
	}

	// Chance of Mutation
	for (int i=0;i<length;i++)
	{
		if (RandFloat() <= MUTATION)
		{
			// Gaussian random perturbation utilizing Box-Mueller Transform
	 		float x1, x2, w, y1, y2;

 	    do {
 		   	x1 = 2.0 * ((rand()%10000)/10000.0) - 1.0;
 	     	x2 = 2.0 * ((rand()%10000)/10000.0) - 1.0;
 	     	w = x1 * x1 + x2 * x2;
   	 	} while ( w >= 1.0 );

   	 	w = sqrt( (-2.0 * log( w ) ) / w );
   	 	y1 = x1 * w;
    	y2 = x2 * w;

			chromosome[i] = chromosome[i] + sqrt(0.15) * y1;
		}
	}  
 	NewBox<T> *offspring = new NewBox<T>(m_WorldID,m_SpaceID,m_StartPosition,true);		// constructor (world,space,x,y,z,in,hid,out,recur)
  offspring->setChromosome(chromosome);
	return offspring;
}

template <class T>
void NewBox<T>::enable(bool enabled)
{
}

template <class T>
void NewBox<T>::setStartPosition(dVector3 start)
{
	// call base function
	Organism<T>::setStartPosition(start);
  dReal x = start[0];
  dReal y = start[1];
  dReal z = start[2];

  // reset sensors and motors
  reset();
  
  // reset the starting position of the bodies
	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	dBodySetPosition (body[1],TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+LEG1[2]/2+GAP));	//right upper leg
	dBodySetPosition (body[2],-TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+LEG1[2]/2+GAP));	// left upper leg
	dBodySetPosition (body[3],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]/2+GAP);	//right foot
	dBodySetPosition (body[4],-TORSO[0]/3-GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]/2+GAP);	//left foot
	dBodySetPosition (body[5],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+LEG3[2]/2+2*GAP);	//right foot
	dBodySetPosition (body[6],-TORSO[0]/3-GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+LEG3[2]/2+2*GAP);	//left foot	

	// reset rotations
  dMatrix3 R;
  dRFromAxisAndAngle (R,0,0,0,0);
  for (int i=0;i<numBody;++i)
		dBodySetRotation(body[i],R);

  // reset linear velocities
  for (int i=0;i<numBody;++i)
		dBodySetLinearVel(body[i],0,0,0);

  // reset angular velocities
  for (int i=0;i<numBody;++i)
		dBodySetAngularVel(body[i],0,0,0);
}

template <class T>
void NewBox<T>::constructBody(const dWorldID world, const dSpaceID space, const dVector3 position)
{
	// Body Setup
	TORSO[0] = 0.7; TORSO[1] = 1.2; TORSO[2] = 0.5;
	LEG1[0] = 0.05; LEG1[1] = 0.05; LEG1[2] = 0.67;
	LEG2[0] = 0.05; LEG2[1] = 0.05; LEG2[2] = 0.67;
	LEG3[0] = 0.35; LEG3[1] = 0.05; LEG3[2] = 0.47;
	GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries
	joint = new dJointID[numJoint];	// allocate space for Joints

	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	//geomGroup = dCreateGeomGroup(space);   // dont use if not needed (slowdown)

	// constants (Mass)
	const dReal MTORSO=2.8;
	const dReal MULEG=0.05;
	const dReal MMLEG=0.05;
	const dReal MLLEG=0.08;

	dMass m;
  dMatrix3 R;
  dReal x = position[0];
  dReal y = position[1];
  dReal z = position[2];

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);
	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	dBodySetPosition (body[1],TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+LEG1[2]/2+GAP));	//right upper leg
	dBodySetPosition (body[2],-TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+LEG1[2]/2+GAP));	// left upper leg
	dBodySetPosition (body[3],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]/2+GAP);	//right foot
	dBodySetPosition (body[4],-TORSO[0]/3-GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]/2+GAP);	//left foot
	dBodySetPosition (body[5],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+LEG3[2]/2+2*GAP);	//right foot
	dBodySetPosition (body[6],-TORSO[0]/3-GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+LEG3[2]/2+2*GAP);	//left foot

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[0],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[0]);

	// upper legs
	dMassSetBox (&m,1,LEG1[0],LEG1[1],LEG1[2]);	// legs
	dMassAdjust (&m,MULEG);	// upper leg weight
	for (int i=1;i<=2;i++)
	{
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
		dGeomSetBody (geom[i],body[i]);
	}

	// middle legs
	dMassSetBox (&m,1,LEG2[0],LEG2[1],LEG2[2]);	// legs
	dMassAdjust (&m,MMLEG);	// middle leg weight
	for (int i=3;i<=4;i++)
	{
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
		dGeomSetBody (geom[i],body[i]);
	}

	// lower legs
	dMassSetBox (&m,1,LEG3[0],LEG3[1],LEG3[2]);	// legs
	dMassAdjust (&m,MLLEG);	// middle leg weight
	for (int i=5;i<=6;i++)
	{
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
		dGeomSetBody (geom[i],body[i]);
	}

	/*for (int i=0;i<5;i++)
		dGeomGroupAdd(geomGroup,geom[i]); */

	/********************************** CREATE JOINTS **********************************/
	// right Leg Socket
	joint[0] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[0],body[0],body[1]);
	dJointSetHingeAnchor(joint[0],TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+GAP/2));
	dJointSetHingeParam(joint[0],dParamLoStop,-1.1);
	dJointSetHingeParam(joint[0],dParamHiStop,0);
	dJointSetHingeParam(joint[0],dParamStopERP,0.2);
	dJointSetHingeParam(joint[0],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[0],dParamBounce,0.3);
	dJointSetHingeParam(joint[0],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[0],dParamVel,0);
	dJointSetHingeParam(joint[0],dParamFMax,0.5);

	// left Leg Socket
	joint[1] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[1],body[0],body[2]);
	dJointSetHingeAnchor(joint[1],-TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+GAP/2));
	dJointSetHingeParam(joint[1],dParamLoStop,-1.1);
	dJointSetHingeParam(joint[1],dParamHiStop,0);
	dJointSetHingeParam(joint[1],dParamStopERP,0.2);
	dJointSetHingeParam(joint[1],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[1],dParamBounce,0.3);
	dJointSetHingeParam(joint[1],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[1],dParamVel,0);
	dJointSetHingeParam(joint[1],dParamFMax,0.5);


	// right middle leg
	joint[2] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[2],body[1],body[3]);
	dJointSetHingeAnchor(joint[2],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+GAP);
	dJointSetHingeAxis(joint[2],1,0,0);
	dJointSetHingeParam(joint[2],dParamLoStop,-1.5);
	dJointSetHingeParam(joint[2],dParamHiStop,0);
	dJointSetHingeParam(joint[2],dParamStopERP,0.3);
	dJointSetHingeParam(joint[2],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[2],dParamBounce,0.3);
	dJointSetHingeParam(joint[2],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[2],dParamVel,0);
	dJointSetHingeParam(joint[2],dParamFMax,0.5);
	// left middle foot
	joint[3] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[3],body[2],body[4]);
	dJointSetHingeAnchor(joint[3],-TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+GAP);
	dJointSetHingeAxis(joint[3],1,0,0);
	dJointSetHingeParam(joint[3],dParamLoStop,-1.5);
	dJointSetHingeParam(joint[3],dParamHiStop,0.0);
	dJointSetHingeParam(joint[3],dParamStopERP,0.3);
	dJointSetHingeParam(joint[3],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[3],dParamBounce,0.3);
	dJointSetHingeParam(joint[3],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[3],dParamVel,0);
	dJointSetHingeParam(joint[3],dParamFMax,0.5);
	// right lower leg
	joint[4] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[4],body[3],body[5]);
	dJointSetHingeAnchor(joint[4],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+(3*GAP)/2);
	dJointSetHingeAxis(joint[4],1,0,0);
	dJointSetHingeParam(joint[4],dParamLoStop,-1.4);
	dJointSetHingeParam(joint[4],dParamHiStop,0.1);
	dJointSetHingeParam(joint[4],dParamStopERP,0.5);
	dJointSetHingeParam(joint[4],dParamStopCFM,0.07);
	dJointSetHingeParam(joint[4],dParamBounce,0.2);
	dJointSetHingeParam(joint[4],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[4],dParamVel,0);
	dJointSetHingeParam(joint[4],dParamFMax,0.5);
	// left lower leg
	joint[5] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[5],body[4],body[6]);
	dJointSetHingeAnchor(joint[5],-TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+(3*GAP)/2);
	dJointSetHingeAxis(joint[5],1,0,0);
	dJointSetHingeParam(joint[5],dParamLoStop,-1.4);
	dJointSetHingeParam(joint[5],dParamHiStop,0.1);
	dJointSetHingeParam(joint[5],dParamStopERP,0.5);
	dJointSetHingeParam(joint[5],dParamStopCFM,0.07);
	dJointSetHingeParam(joint[5],dParamBounce,0.2);
	dJointSetHingeParam(joint[5],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[5],dParamVel,0);
	dJointSetHingeParam(joint[5],dParamFMax,0.5);

	dRFromAxisAndAngle (R, 0,0,0,0);
	for (int i=0;i<7;i++)
		dBodySetRotation(body[i], R);
}
